package be.ASPL.presentationlayer;

import java.io.*;
import java.util.*;
import java.util.Map.Entry;
import java.util.logging.Logger;

import be.SIRAPRISE.util.MyMessageFormat;
import be.SIRAPRISE.util.MyReadOnlySet;
import be.SIRAPRISE.util.MyString;

/**
 * The Architecture singleton is instantiated when the first Architecture Servlet is initialized. That servlet maintains a reference to the Architecture object, thus ensuring that the Architecture object cannot be finalized as long as there is any Architecture Servlet that is still accessible to the JVM. The Architecture object maintains references to several pools, such as the user session pool. Doing so ensures us that all the resources needed by the WAAR architecture will exist and be consistent as long as any Architecture Servlet is kept active by the Application server.
 * 
 * @author Erwin
 */
public class Architecture {

	/**
	 * The AttributeTypePool is the map holding all the Presentation Type objects (mapped to their name).
	 * 
	 * @author Erwin
	 */

	public static class AttributeTypePool {

		/**
		 * The instance of the pool
		 */
		static AttributeTypePool attributeTypePoolInstance = new AttributeTypePool();

		/**
		 * The pool of Types for presentation attributes
		 */
		private HashMap<String, PresentationType<?>> pool = new HashMap<String, PresentationType<?>>();

		/**
		 * Creates the pool
		 */
		private AttributeTypePool ( ) {

		}

		/**
		 * Attempts to instantiate an object of a PresentationAttributeType class
		 * 
		 * @param typeName
		 *            The (fully qualified) name of a PresentationAttributeType that is to be instantiated
		 * @param presentationFunctionClass
		 *            The Class object of the presentation function for which the type is needed. This object acts as a handle to its classLoader, which might be needed to gain access to package-defined presentation types (such as e.g. enumeration types)
		 * @return An object of the Classname that is constructed by concatenating the given package name and type Name
		 * @throws NotAPresentationAttributeTypeException
		 */
		private static PresentationType<?> loadPresentationTypeFromPackage (String typeName, Class<? extends OnLinePresentationFunction> presentationFunctionClass) throws NotAPresentationAttributeTypeException {
			Object object;
			try {
				//				object = Class.forName(packageName + "." + typeName).newInstance(); //$NON-NLS-1$
				//				object = getClass().getClassLoader().loadClass(packageName + "." + typeName).newInstance(); //$NON-NLS-1$
				object = presentationFunctionClass.getClassLoader().loadClass(typeName).newInstance();
			} catch (InstantiationException e) {
				throw new NotAPresentationAttributeTypeException(e);
			} catch (IllegalAccessException e) {
				throw new NotAPresentationAttributeTypeException(e);
			} catch (ClassNotFoundException e) {
				throw new NotAPresentationAttributeTypeException(e);
			}
			if (object instanceof PresentationType) {
				return (PresentationType<?>) object;
			} else {
				throw new NotAPresentationAttributeTypeException();
			}
		}

		/**
		 * Gets the PresentationType object associated with a name. The method first looks in the pool for a Presentation type of the given name. If no such type is present in the pool, then an attempt is made to instantiate a Class of the given typename, prefixed with the java package name of the Architecture. If that fails, an attempt is made to instantiate a Class of the given typename, prefixed with the given application package name. The Class must be a subClass of PresentationAttributeType.
		 * 
		 * @param typeName
		 *            The name of the attribute Type to be obtained
		 * @param presentationFunctionClass
		 *            The Class object of the presentation function for which the type is needed. This object acts as a handle to its classLoader, which might be needed to gain access to package-defined presentation types (such as e.g. enumeration types)
		 * @return The requested AttributeType object
		 * @throws FailedToCreatePresentationTypeException
		 *             if the Type object for the given name cannot be instantiated
		 */
		PresentationType<?> get (String typeName, Class<? extends OnLinePresentationFunction> presentationFunctionClass) throws FailedToCreatePresentationTypeException {
			synchronized (pool) {
				PresentationType<?> presentationType;
				if (pool.containsKey(typeName)) {
					presentationType = pool.get(typeName);
				} else {
					try {
						presentationType = loadPresentationTypeFromPackage(typeName, presentationFunctionClass);
					} catch (NotAPresentationAttributeTypeException e) {
						try {
							presentationType = loadPresentationTypeFromPackage(typeName, presentationFunctionClass);
						} catch (NotAPresentationAttributeTypeException e1) {
							e1.printStackTrace(System.err);
							throw new FailedToCreatePresentationTypeException(e1, typeName);
						}
					}
					pool.put(typeName, presentationType);
				}
				return presentationType;
			}
		}

		/**
		 * Gets the presentationtype object corresponding to the given PresentationType class
		 * 
		 * @param typeClass
		 *            The PresentationType class for which a corresponding PresentationType object is to be returned
		 * @return the presentationtype object corresponding to the given PresentationType class
		 * @throws FailedToCreatePresentationTypeException
		 *             -
		 */
		public PresentationType<?> get (Class<? extends PresentationType<?>> typeClass) throws FailedToCreatePresentationTypeException {
			synchronized (pool) {
				PresentationType<?> presentationType;
				final String typeName = typeClass.getName();
				if (pool.containsKey(typeName)) {
					presentationType = pool.get(typeName);
				} else {
					try {
						presentationType = typeClass.newInstance();
					} catch (InstantiationException e) {
						throw new FailedToCreatePresentationTypeException(e, typeName);
					} catch (IllegalAccessException e) {
						throw new FailedToCreatePresentationTypeException(e, typeName);
					}
					pool.put(typeName, presentationType);
				}
				return presentationType;
			}
		}

		/**
		 * Gets the pool
		 * 
		 * @return the pool
		 */
		public HashMap<String, PresentationType<?>> getPool ( ) {
			return pool;
		}

		/**
		 * Removes a Presentation Type from the pool
		 * 
		 * @param typeName
		 *            The name of the type whose entry is to be removed from the pool
		 */
		public void remove (String typeName) {
			synchronized (pool) {
				pool.remove(typeName);
			}
		}
	}

	/**
	 * 
	 */
	private static final String ARCHITECTURE_HTMLTEMPLATE = "architecture.htmltemplate"; //$NON-NLS-1$

	/**
	 * 
	 */
	private static final Locale architectureLocale = new Locale("EN"); //$NON-NLS-1$

	/**
	 * 
	 */
	private static final String ASPL = "ASPL"; //$NON-NLS-1$

	/**
	 * 
	 */
	private static final String ASPLADMINISTRATOR_PWD = "AsplAdministrator"; //$NON-NLS-1$

	/**
	 * 
	 */
	private static final String ASPLADMINISTRATOR_USERID = "aspladministrator"; //$NON-NLS-1$

	/**
	 * The map holding the registered WAAR users and properties
	 */
	private static HashMap<String, UserProperties> asplUsers;

	/**
	 * The pool of Presentation layer attributes
	 */
	private static AttributeTypePool attributeTypePool;

	/**
	 * The instance of the Architecture context
	 */
	private static Architecture instance = new Architecture();

	/**
	 * The Map of Presentation architecture fields
	 */
	private static PresentationArchitectureFieldMap presentationArchitectureFieldMap = PresentationArchitectureFieldMap.getInstance();

	/**
	 * The map of supported languages
	 */
	private static HashMap<String, String> supportedLanguages;

	/**
	 * 
	 */
	private static final String USERDATABASEFILENAMESUFFIX = ".udb"; //$NON-NLS-1$

	/**
	 * The actual map holding all the currently active user sessions
	 */
	private static HashMap<Long, UserSession> userSessions = new HashMap<Long, UserSession>();

	/**
	 * The common html for all architecture output
	 */
	static String architectureHtmlTemplate;

	/**
	 * The home directory of the WAAR application
	 */
	static String asplHomeDirectory;

	/**
	 * 
	 */
	public static final String ASPLADMINISTRATOR_ROLENAME = "aspladmin"; //$NON-NLS-1$

	/**
	 * @param applicationName
	 */
	private static synchronized void initUsersMap (String applicationName) {
		if (asplUsers == null) {
			boolean writeUDBFile = false;
			Properties p = new Properties();
			File userDatabaseFile = new File(asplHomeDirectory + File.separatorChar + applicationName + USERDATABASEFILENAMESUFFIX);
			if (userDatabaseFile.exists()) {
				try (FilterInputStream userDatabaseFileInputStream = new UDBInputStream(userDatabaseFile)) {
					p.load(userDatabaseFileInputStream);
				} catch (IOException e) {

				}
			} else {
				p.setProperty(ASPLADMINISTRATOR_USERID, ASPLADMINISTRATOR_PWD);
				writeUDBFile = true;
			}

			asplUsers = new HashMap<String, UserProperties>();
			for (Entry<Object, Object> me : p.entrySet()) {
				asplUsers.put((String) me.getKey(), new UserProperties((String) me.getValue() + UserProperties.SEP + "true" + UserProperties.SEP + "[]" + UserProperties.SEP + "EN" + UserProperties.SEP + ASPLADMINISTRATOR_ROLENAME)); //$NON-NLS-1$//$NON-NLS-2$ //$NON-NLS-3$
			}
			if (writeUDBFile) {
				writeUDBFile(applicationName);
			}
		}
	}

	/**
	 * @param applicationName
	 * 
	 */
	private static synchronized void writeUDBFile (String applicationName) {
		File userDatabaseFile = new File(asplHomeDirectory + File.separatorChar + applicationName + USERDATABASEFILENAMESUFFIX);
		userDatabaseFile.getParentFile().mkdirs();
		try (DataOutputStream userDatabaseOutputStream = new DataOutputStream(new UDBOutputStream(userDatabaseFile))) {
			for (Entry<String, UserProperties> me : asplUsers.entrySet()) {
				userDatabaseOutputStream.writeBytes(me.getKey() + '=' + me.getValue().toString() + "\r\n"); //$NON-NLS-1$
			}
		} catch (IOException e1) {
			Logger.getLogger(Architecture.class.getName()).severe(e1.getClass().getName() + ' ' + e1.getMessage());
		}
	}

	/**
	 * @param applicationName
	 * @param userID
	 * @param userPassword
	 * @param newPassword
	 *            the new password, if no password change was requested then null or the empty string
	 * @param confirmNewPassword
	 *            the new password, if no password change was requested then null or the empty string
	 * @throws InvalidPasswordException
	 * @throws UnknownUserIDException
	 * @throws InvalidNewPasswordException
	 * @throws PasswordUnconfirmedException
	 * @throws MaximumInvalidPasswordsForUserExceededException
	 * @throws UserIDDisabledException
	 */
	static void checkPassword (String applicationName, String userID, String userPassword, String newPassword, String confirmNewPassword) throws InvalidPasswordException, UnknownUserIDException, InvalidNewPasswordException, PasswordUnconfirmedException, MaximumInvalidPasswordsForUserExceededException, UserIDDisabledException {
		initUsersMap(applicationName);
		synchronized (asplUsers) {
			if (asplUsers.containsKey(userID)) {
				UserProperties userProperties = asplUsers.get(userID);
				if (userProperties.isDisabled()) {
					throw new UserIDDisabledException(userID);
				}
				if (userProperties.getPassword().equals(userPassword)) {
					if (userProperties.isPasswordToBeChangedAtNextLogon() || (newPassword != null && newPassword.length() > 0)) {
						if (newPassword.length() < 8 || newPassword.equals(userPassword)) {
							throw new InvalidNewPasswordException();
						}
						if (newPassword.equals(confirmNewPassword)) {
							asplUsers.put(userID, new UserProperties(newPassword + UserProperties.SEP + "false" + UserProperties.SEP + userProperties.getSuspiciousLogonAttemptTimes() + UserProperties.SEP + userProperties.getLanguage() + UserProperties.SEP + MyString.asCSV(userProperties.getRoles()))); //$NON-NLS-1$
							writeUDBFile(applicationName);
						} else {
							throw new PasswordUnconfirmedException();
						}
					}
				} else {
					try {
						userProperties.registerSuspiciousLogonAttempt();
						throw new InvalidPasswordException();
					} catch (MaximumInvalidPasswordsExceededException e) {
						throw new MaximumInvalidPasswordsForUserExceededException(userID);
					} finally {
						writeUDBFile(applicationName);
					}
				}
			} else {
				throw new UnknownUserIDException(userID);
			}
		}
	}

	/**
	 * Gets the String object indicated by the given String
	 * 
	 * @param userLanguage
	 *            a valid 2-char ISO language code
	 * @return The String corresponding to the given language code
	 * @throws InvalidAttributeValueException
	 */
	static String checkUserLanguage (String userLanguage) throws InvalidAttributeValueException {
		String userLanguage_uc = userLanguage.toUpperCase();
		if (supportedLanguages.containsKey(userLanguage_uc)) {
			return userLanguage_uc;
		}
		throw new InvalidAttributeValueException(userLanguage, "Is not a supported language"); //$NON-NLS-1$
	}

	/**
	 * Removes a session from the pool
	 * 
	 * @param sessionID
	 *            The identification number of the session to be removed
	 */
	static void dropUserSession (long sessionID) {
		synchronized (userSessions) {
			userSessions.remove(new Long(sessionID));
		}
	}

	/**
	 * Obtains the value for some field of the Presentation Architecture, as it applies to some given user function.
	 * 
	 * @param fieldName
	 *            The name of the Presentation Architecture field whose value is to be obtained
	 * @param userFunction
	 *            The user Function for which the field value is requested. (This is relevant for fields such as the function's application name, the system date (needs formatting info that resides in the function's user data).)
	 * @return The value of the architecture field in string format
	 * @throws NotAnArchitectureFieldException
	 *             if the given fieldName is not a known fieldName of the Presentation architecture
	 */
	static String getArchitectureFieldValue (String fieldName, OnLinePresentationFunction userFunction) throws NotAnArchitectureFieldException {
		if (presentationArchitectureFieldMap.containsKey(fieldName)) {
			ArchitectureValueGetter valueGetter = presentationArchitectureFieldMap.get(fieldName);
			return valueGetter.getValue(userFunction);
		} else {
			throw new NotAnArchitectureFieldException(fieldName);
		}
	}

	/**
	 * Gets the AttributeType object associated with a name.
	 * 
	 * @param typeName
	 *            The name of the attribute Type to be obtained
	 * @param presentationFunctionClass
	 *            The Class object of the presentation function for which the type is needed. This object acts as a handle to its classLoader, which might be needed to gain access to package-defined presentation types (such as e.g. enumeration types)
	 * @return The requested AttributeType object
	 * @throws FailedToCreatePresentationTypeException
	 */
	static PresentationType<?> getAttributeType (String typeName, Class<? extends OnLinePresentationFunction> presentationFunctionClass) throws FailedToCreatePresentationTypeException {
		return attributeTypePool.get(typeName, presentationFunctionClass);
	}

	/**
	 * @param typeClass
	 * @return
	 * @throws FailedToCreatePresentationTypeException
	 */
	static PresentationType<?> getAttributeType (Class<? extends PresentationType<?>> typeClass) throws FailedToCreatePresentationTypeException {
		return attributeTypePool.get(typeClass);
	}

	/**
	 * Gets The instance of the Architecture
	 * 
	 * @return The instance of the Architecture
	 */
	static Architecture getInstance ( ) {
		return instance;
	}

	/**
	 * @param resourceName
	 * @param presentationFunctionClass
	 * @return
	 * @throws ResourceNotFoundException
	 */
	static InputStream getResourceInputStream (String resourceName, Class<?> presentationFunctionClass) throws ResourceNotFoundException {
		return resourceInputStream(resourceName, presentationFunctionClass);
	}

	/**
	 * @param resourceName
	 * @param presentationFunctionClass
	 * @return
	 * @throws ResourceNotFoundException
	 */
	static BufferedReader getResourceReader (String resourceName, Class<?> presentationFunctionClass) throws ResourceNotFoundException {
		return resourceReader(resourceName, presentationFunctionClass);
	}

	/**
	 * @param applicationName
	 * @param userID
	 * @return
	 */
	static MyReadOnlySet<String> getUserRoles (String applicationName, String userID) {
		initUsersMap(applicationName);
		synchronized (asplUsers) {
			if (asplUsers.containsKey(userID)) {
				return asplUsers.get(userID).getRoles();
			} else {
				return new MyReadOnlySet<>(new HashSet<String>());
			}
		}
	}

	/**
	 * Gets the user session pool. Method should never be invoked. The Architecture just needs to hold a reference to this pool in order to prevent it from being finalized.
	 * 
	 * @param sessionID
	 *            The identification number of the user session to retrieve
	 * @return the requested user session, or null if such a session does not exist
	 */
	static UserSession getUserSession (long sessionID) {
		synchronized (userSessions) {
			return userSessions.get(new Long(sessionID));
		}
	}

	/**
	 * @return
	 */
	static Map<String, String> supportedLanguages ( ) {
		return supportedLanguages;
	}

	/**
	 * @param applicationName
	 *            -
	 * @param userID
	 *            -
	 * @param newLanguage
	 *            -
	 * @param rolesCSV
	 *            -
	 * @throws UnknownUserIDException
	 *             -
	 */
	public static void checkAndSetLanguagePlusRoles (String applicationName, String userID, String newLanguage, String rolesCSV) throws UnknownUserIDException {
		initUsersMap(applicationName);
		synchronized (asplUsers) {
			if (asplUsers.containsKey(userID)) {
				UserProperties userProperties = asplUsers.get(userID);
				if (supportedLanguages.containsKey(newLanguage)) {
					asplUsers.put(userID, new UserProperties(userProperties.getPassword() + UserProperties.SEP + userProperties.isPasswordToBeChangedAtNextLogon() + UserProperties.SEP + userProperties.getSuspiciousLogonAttemptTimes() + UserProperties.SEP + newLanguage + UserProperties.SEP + rolesCSV));
					writeUDBFile(applicationName);
				}
			} else {
				throw new UnknownUserIDException(userID);
			}
		}
	}

	/**
	 * Gets The map holding the registered WAAR users and properties
	 * 
	 * @param applicationName
	 *            -
	 * 
	 * @return The map holding the registered WAAR users and properties
	 */
	public static HashMap<String, UserProperties> getAsplUsers (String applicationName) {
		initUsersMap(applicationName);
		return asplUsers;
	}

	/**
	 * Gets the AttributeTypePool
	 * 
	 * @return the AttributeTypePool
	 */
	public static AttributeTypePool getAttributeTypePool ( ) {
		return attributeTypePool;
	}

	/**
	 * Gets The default locale of the installation
	 * 
	 * @return The default locale of the installation
	 */
	public static Locale getDefaultLocale ( ) {
		return architectureLocale;
	}

	/**
	 * Gets the UserProperties object holding the properties of the given user ID
	 * 
	 * @param applicationName
	 *            -
	 * @param userID
	 *            The user ID whose properties are to be obtained
	 * 
	 * @return The UserProperties object holding the properties of the given user ID
	 * @throws UserIDUnknownException
	 *             -
	 */
	public static UserProperties getUserProperties (String applicationName, String userID) throws UserIDUnknownException {
		initUsersMap(applicationName);
		synchronized (asplUsers) {
			if (asplUsers.containsKey(userID.toLowerCase())) {
				return asplUsers.get(userID.toLowerCase());
			}
		}
		throw new UserIDUnknownException(userID);
	}

	/**
	 * @param applicationName
	 *            -
	 * @param userID
	 *            -
	 * @param password
	 *            -
	 * @param userLanguage
	 *            -
	 * @param rolesCSV
	 *            -
	 * @throws UserAlreadyExistsException
	 *             -
	 */
	public static void newWaarUser (String applicationName, String userID, String password, String userLanguage, String rolesCSV) throws UserAlreadyExistsException {
		String lowerCaseUserID = userID.toLowerCase();
		initUsersMap(applicationName);
		synchronized (asplUsers) {
			if (asplUsers.containsKey(lowerCaseUserID)) {
				throw new UserAlreadyExistsException(userID);
			}

			asplUsers.put(lowerCaseUserID, new UserProperties(password + UserProperties.SEP + "true" + UserProperties.SEP + "[]" + UserProperties.SEP + userLanguage + UserProperties.SEP + rolesCSV)); //$NON-NLS-1$ //$NON-NLS-2$
			writeUDBFile(applicationName);
		}
	}

	/**
	 * @param applicationName
	 *            -
	 * @param userID
	 *            -
	 * @throws AdministratorCannotBeRemovedException
	 *             -
	 * @throws UserIDUnknownException
	 *             ---
	 */
	public static void removeWaarUser (String applicationName, String userID) throws AdministratorCannotBeRemovedException, UserIDUnknownException {
		String lowerCaseUserID = userID.toLowerCase();
		initUsersMap(applicationName);
		synchronized (asplUsers) {
			if (!asplUsers.containsKey(lowerCaseUserID)) {
				throw new UserIDUnknownException(userID);
			}
			if (lowerCaseUserID.equals(ASPLADMINISTRATOR_USERID)) {
				throw new AdministratorCannotBeRemovedException();
			}

			asplUsers.remove(lowerCaseUserID);
			writeUDBFile(applicationName);
		}
	}

	/**
	 * The last assigned session ID
	 */
	private long userSessionIDLastAssigned;

	/**
	 * The Architecture is instantiated at the time the first Architecture Servlet is initialized. It will not be destroyed as long as there are any Architecture Servlets "active" in the JVM in which the Application server is running.
	 */
	private Architecture ( ) {

		asplHomeDirectory = System.getProperty("user.home") + File.separatorChar + ASPL; //$NON-NLS-1$
		new File(asplHomeDirectory).mkdirs();

		attributeTypePool = AttributeTypePool.attributeTypePoolInstance;

		StringBuffer htmlTemplateText = new StringBuffer(8192);
		String w;

		try {
			
			try (BufferedReader in = resourceReader(ARCHITECTURE_HTMLTEMPLATE, Architecture.class);) {
				while ((w = in.readLine()) != null) {
					htmlTemplateText.append(w);
				}
				architectureHtmlTemplate = htmlTemplateText.toString();
			} catch (IOException e1) {
				e1.printStackTrace(System.err);
				architectureHtmlTemplate = "<HTML><BODY>" + MyMessageFormat.format(I18N.getString("Architecture.ArchitectureTemplateError"), new String[] { e1.getClass().getName(), e1.getMessage() }) + "</BODY></HTML>"; //$NON-NLS-1$ //$NON-NLS-2$  //$NON-NLS-3$
			}
		} catch (ResourceNotFoundException e3) {
			e3.printStackTrace(System.err);
			architectureHtmlTemplate = "<HTML><BODY>" + MyMessageFormat.format(I18N.getString("Architecture.ArchitectureHTMLNotFound"), new String[] { ARCHITECTURE_HTMLTEMPLATE }) + "</BODY></HTML>"; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		}

		// Set up the languages
		supportedLanguages = new HashMap<String, String>();
		supportedLanguages.put("NL", "Nederlands"); //$NON-NLS-1$//$NON-NLS-2$
		supportedLanguages.put("FR", "Fran\u00E7ais"); //$NON-NLS-1$//$NON-NLS-2$
		supportedLanguages.put("DE", "Deutsch"); //$NON-NLS-1$//$NON-NLS-2$
		supportedLanguages.put("EN", "English"); //$NON-NLS-1$//$NON-NLS-2$
		supportedLanguages.put("ES", "Espa\u00F1ol"); //$NON-NLS-1$//$NON-NLS-2$
		supportedLanguages.put("PT", "Portugues"); //$NON-NLS-1$//$NON-NLS-2$
		supportedLanguages.put("IT", "Italiano"); //$NON-NLS-1$ //$NON-NLS-2$
		supportedLanguages.put("NO", "Norsk"); //$NON-NLS-1$//$NON-NLS-2$
		supportedLanguages.put("DK", "Dansk"); //$NON-NLS-1$//$NON-NLS-2$
		supportedLanguages.put("SE", "Svenska"); //$NON-NLS-1$ //$NON-NLS-2$
		supportedLanguages.put("FI", "Suomen"); //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * @param resourceName
	 * @param resourceClass
	 * @return
	 * @throws ResourceNotFoundException
	 */
	private static InputStream resourceInputStream (String resourceName, Class<?> resourceClass) throws ResourceNotFoundException {
		InputStream resourceInputStream = resourceClass.getResourceAsStream(resourceName);
		if (resourceInputStream == null) {
			resourceInputStream = resourceClass.getResourceAsStream("/" + resourceName); //$NON-NLS-1$
			if (resourceInputStream == null) {
				// String presentationFileName = Architecture.waarHomeDirectory + File.separatorChar + resourceClass + (File.separatorChar + Architecture.WEB_CONTENT + File.separatorChar + resourceName);
				// File presentationFile = new File(presentationFileName);
				// try {
				// in = new BufferedReader(new FileReader(presentationFile));
				// } catch (FileNotFoundException e2) {
				throw new ResourceNotFoundException();
				// }
			}
		}
		return resourceInputStream;
	}

	/**
	 * @param resourceName
	 * @param resourceClass
	 * @return
	 * @throws ResourceNotFoundException
	 */
	private static BufferedReader resourceReader (String resourceName, Class<?> resourceClass) throws ResourceNotFoundException {
		return new BufferedReader(new InputStreamReader(resourceInputStream(resourceName, resourceClass)));
	}

	/**
	 * Gets a new UserSession
	 * 
	 * @param userSessionState
	 * 
	 * @return a new UserSession
	 */
	UserSession newUserSession (String userSessionState) {
		UserSession userSession;
		if (userSessionState == null) {
			synchronized (userSessions) {
				userSessionIDLastAssigned++;
				userSession = new UserSession(userSessionIDLastAssigned);
				userSessions.put(new Long(userSessionIDLastAssigned), userSession);
			}
		} else {
			// userSession = getSessionFrom(userSessionState);
			throw new UnsupportedOperationException("Stateless operation not yet supported."); //$NON-NLS-1$
		}
		return userSession;
	}
}
