package be.ASPL.presentationlayer;

import java.io.*;
import java.util.Collection;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.Properties;
import java.util.Map.Entry;

import javax.servlet.*;
import javax.servlet.annotation.MultipartConfig;
import javax.servlet.annotation.WebServlet;
import javax.servlet.http.*;

/**
 * @version 1.0
 * @author Erwin
 */
public abstract class HTTPArchitectureServlet extends HttpServlet {

	/**
	 * 
	 */
	private static final String defaultURLSuffixFromAnnotation = ".wfn"; //$NON-NLS-1$

	/**
	 * 
	 */
	private static final long serialVersionUID = -7925812282659471814L;

	/**
	 * @param userSession
	 * @return
	 */
	private static String getUserSessionRequestParameter (UserSession userSession) {
		if (userSession.getSessionID() >= 0) {
			return "USERSESSIONID=" + userSession.getSessionID(); //$NON-NLS-1$
		} else {
			return "USERSESSIONSTATE=" + userSession.encodeState(); //$NON-NLS-1$
		}
	}

	/**
	 * @param req
	 * @return
	 */
	static Map<String, String> getParameterValuesMap (HttpServletRequest req) {
		Map<String, String> m = new HashMap<String, String>();
		for (Entry<String, String[]> me : req.getParameterMap().entrySet()) {
			if (me.getValue().length > 0) {
				m.put(me.getKey(), me.getValue()[0]);
			}
		}
		return m;
	}

	/**
	 * Injects a property value for the given property name in a given user Properties object. The injection is done only if the given value is non-null and the given Properties object does not yet have an entry [that is non-null] for the given property name.
	 * 
	 * @param applicationPropertiesObject
	 *            The Properties object to inject a value into
	 * @param propertyName
	 *            The property name to inject
	 * @param propertyValue
	 *            The property value to inject
	 */
	protected static void injectProperty (Properties applicationPropertiesObject, String propertyName, String propertyValue) {
		if (applicationPropertiesObject.getProperty(propertyName) == null && propertyValue != null) {
			applicationPropertiesObject.setProperty(propertyName, propertyValue);
		}
	}

	/**
	 * Injects all property values found for the given Properties object in the servlet's ServletContext (that is, the ServletContext's set of InitParameters), into that given Properties object. A servlet's ServletContext's InitParameter is suitable for injecting into the given Properties object if and only if its full ParameterName matches (that is, is longer than and its starting sequence is equal to) the fully-qualified classname of the given Properties object. The property name injected is what follows after the mandatory initial '.' token of the remainder of the InitParameter name.
	 * 
	 * @param applicationPropertiesObject
	 *            The Properties object to inject values from the servlet's ServletContext into
	 */
	protected void injectProperties (Properties... applicationPropertiesObject) {
		final ServletContext context = getServletContext();
		final Enumeration<String> contextInitParameterNames = context.getInitParameterNames();
		while (contextInitParameterNames.hasMoreElements()) {
			final String contextInitParameterName = contextInitParameterNames.nextElement();
			for (int i = 0; i < applicationPropertiesObject.length; i++) {
				final String propertiesObjectClassName = applicationPropertiesObject[i].getClass().getName();
				if (contextInitParameterName.length() > propertiesObjectClassName.length() && contextInitParameterName.startsWith(propertiesObjectClassName + '.')) {
					final String propertyName = contextInitParameterName.substring(propertiesObjectClassName.length() + 1);
					if (applicationPropertiesObject[i].getProperty(propertyName) == null) {
						applicationPropertiesObject[i].setProperty(propertyName, context.getInitParameter(contextInitParameterName));
					}
				}
			}
		}
	}

	/**
	 * the Architecture object (instance). This reference is maintained to ensure that the Architecture singleton remains in the JVM at least as long as any HTTPArchitectureServlet is maintained by the Application Server. It should not normally be referred to by any method.
	 */
	private Architecture architecture;

	/**
	 * 
	 */
	private String urlSuffixFromAnnotation;

	/**
	 * @return
	 */
	private String computeURLSuffixFromAnnotation ( ) {
		WebServlet ann = getClass().getAnnotation(WebServlet.class);
		if (ann == null) {
			return defaultURLSuffixFromAnnotation;
		} else {
			final String[] urlPatterns = ann.urlPatterns();
			if (urlPatterns == null || urlPatterns.length < 1) {
				return defaultURLSuffixFromAnnotation;
			} else {
				for (int i = 0; i < urlPatterns.length; i++) {
					String urlPattern = urlPatterns[i];
					if (urlPattern.startsWith("*.")) { //$NON-NLS-1$
						return urlPattern.substring(1);
					}
				}
				return defaultURLSuffixFromAnnotation;
			}
		}
	}

	/**
	 * @param req
	 * @return
	 * @throws ServletException
	 */
	private Collection<Part> getParts (HttpServletRequest req) throws ServletException {
		try {
			return this.getClass().getAnnotation(MultipartConfig.class) == null ? new LinkedList<Part>() : req.getParts();
		} catch (IOException e) {
			throw new ServletException(e);
		}
	}

	/**
	 * @return
	 */
	private String getURLSuffixFromAnnotation ( ) {
		return (urlSuffixFromAnnotation == null) ? (urlSuffixFromAnnotation = computeURLSuffixFromAnnotation()) : urlSuffixFromAnnotation;
	}

	/**
	 * Processes a request.
	 * 
	 * @param req
	 *            The request
	 * @param resp
	 *            The object through which the response is returned
	 * @throws ServletException
	 */
	private final void process (HttpServletRequest req, HttpServletResponse resp) throws ServletException {
		resp.setContentType("text/html;charset=" + ArchitectureValueGetterForCharSet.UTF8); //$NON-NLS-1$
		resp.setCharacterEncoding(ArchitectureValueGetterForCharSet.UTF8);
		//		resp.setHeader("accept-charset", ArchitectureValueGetterForCharSet.UTF8); //$NON-NLS-1$

		if (req.getCharacterEncoding() == null) {
			try {
				req.setCharacterEncoding(ArchitectureValueGetterForCharSet.UTF8);
			} catch (UnsupportedEncodingException e1) {
				e1.printStackTrace(System.err);
			}
		}

		String applicationName = req.getContextPath().substring(1); // "/ASPL" ===> "ASPL"
		String functionName = req.getServletPath().substring(1);
		final String urlSuffixFromAnnotation2 = getURLSuffixFromAnnotation();
		if (functionName.endsWith(urlSuffixFromAnnotation2)) {
			functionName = functionName.substring(0, functionName.length() - urlSuffixFromAnnotation2.length());
		}

		// Get the USERSESSIONID / USERSESSIONTOKEN combo from the request,
		// if not present create a new UserSession object, else
		// use it to fetch the UserSession object from the Sessions pool,
		// and verify its token against the TOKEN from the request
		String userSessionIDText = req.getParameter("USERSESSIONID"), sessionToken = req.getParameter("USERSESSIONTOKEN"); //$NON-NLS-1$ //$NON-NLS-2$
		UserSession userSession;
		if (userSessionIDText == null) {
			userSession = architecture.newUserSession(req.getParameter("USERSESSIONSTATE")); //$NON-NLS-1$
		} else {
			long userSessionID = Long.parseLong(userSessionIDText);
			userSession = Architecture.getUserSession(userSessionID);
			if (userSession == null) {
				log("User Session no longer active : " + req.getRequestURL()); //$NON-NLS-1$
				req.getSession().invalidate();
				throw new ServletException("User Session no longer active."); //$NON-NLS-1$
			}
			if (Long.parseLong(sessionToken) != userSession.getSessionVerificationToken()) {
				log("Session token verification failed : " + req.getRequestURL()); //$NON-NLS-1$
				req.getSession().invalidate();
				Architecture.dropUserSession(userSession.getSessionID());
				throw new ServletException("Session token verification failed."); //$NON-NLS-1$
			}
		}

		try {
			UserData userData = userSession.getUserdata();
			if (userData == null) {
				// this will be the case if the UserSession was just newly created because the request URL did not hold a session identifier
				resp.sendRedirect(resp.encodeRedirectURL("/" + applicationName + "/Logon?" + getUserSessionRequestParameter(userSession) + "&return=" + req.getRequestURL())); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
			} else {
				// this will be the case if the request URL did hold a session identifier
				// In this case, get the "Action" parameter from the request (which identifies the presentation button pressed)
				OnLinePresentationFunction userFunction = null;
				String action = req.getParameter("Action"); //$NON-NLS-1$
				if (action == null) {
					// Then the dialog is to be initialized or resumed
					userFunction = processInitOrResume(req, applicationName, functionName, userSession, userData);
				} else {
					// Then the Action identified (e.g. OK, CANCEL, HELP) is to be carried out
					userFunction = processAction(req, resp, userSession, action);
				}

				if (userFunction != null) {
					// There is a function whose presentation must be shown
					final String memorizeData = userFunction.getMemorizeData();
					if (memorizeData != null) {
						final Cookie cookie = new Cookie(userFunction.getClass().getSimpleName(), memorizeData);
						cookie.setPath(req.getContextPath() + req.getServletPath());
						resp.addCookie(cookie);
					}
					resp.getWriter().write(userFunction.getHTMLOutput());
				} else {
					// This well be the case if we have CANCELled from the application's main menu, i.e. there is no application function to return to
					Architecture.dropUserSession(userSession.getSessionID());
					resp.sendRedirect(resp.encodeRedirectURL("/" + applicationName)); //$NON-NLS-1$
				}
			}
		} catch (IOException e) {
			log(e.getMessage(), e);
			req.getSession().invalidate();
			Architecture.dropUserSession(userSession.getSessionID());
			throw new ServletException(e);
		} catch (RuntimeException e) {
			log(e.getMessage(), e);
			req.getSession().invalidate();
			Architecture.dropUserSession(userSession.getSessionID());
			throw new ServletException(e);
		}
	}

	/**
	 * A Request has been received that contained an Action= parameter. This means that some action button has been pushed in the current presentation function.
	 * 
	 * @param req
	 *            The request
	 * @param resp
	 *            The object through which the response is returned
	 * @param userSession
	 *            The user session in whose context the function is invoked
	 * @param action
	 *            The value of the Action= parameter
	 * @return The presentation function that will write the output
	 * @throws ServletException
	 */
	@SuppressWarnings("deprecation")
	private final OnLinePresentationFunction processAction (HttpServletRequest req, HttpServletResponse resp, UserSession userSession, String action) throws ServletException {
		final PresentationFunctionStack functionStack = userSession.getFunctionStack();
		// action=... ; process requested action
		OnLinePresentationFunction userFunction = (OnLinePresentationFunction) functionStack.getLast();
		try {
			switch (action.toUpperCase()) {
				case "OK": { //$NON-NLS-1$
					// OK on the HELP presentation means leave the help and redisplay the function's real user presentation
					// OK on the real function presentation means "process the entered data"
					if (userFunction.isHelpPresentationPoppedUp()) {
						userFunction.popoffPresentation();
						userFunction.setHelpPresentationPoppedUp(false);
					} else {
						userFunction.fillInPresentationValuesIgnoringDefaults(getParameterValuesMap(req), getParts(req));
						userFunction.setPresentationMessage(""); //$NON-NLS-1$
						userFunction.doPresentationLogic();
						userFunction.endPresentationLogic();
						if (userFunction.leaveFunctionOnExit()) {
							userFunction = processReturn(req, resp, userSession, functionStack);
						} else {
							// stay on same function after OK ===> do nothing
						}
					}
					break;
				}
				case "CANCEL": { //$NON-NLS-1$
					if (userFunction.isHelpPresentationPoppedUp()) {
						userFunction.popoffPresentation();
						userFunction.setHelpPresentationPoppedUp(false);
					} else {
						userFunction.doCancel();
						userFunction = processReturn(req, resp, userSession, functionStack);
					}
					break;
				}
				case "HELP": { //$NON-NLS-1$
					if (!userFunction.isHelpPresentationPoppedUp()) {
						HelpPresentation helpPresentationAnnotation = userFunction.getClass().getAnnotation(HelpPresentation.class);
						if (helpPresentationAnnotation != null) {
							userFunction.popupPresentation(helpPresentationAnnotation.helpPresentationClass());
						} else {
							userFunction.popupPresentation(userFunction.getClass().getSimpleName() + "Help"); //$NON-NLS-1$
						}
						userFunction.setHelpPresentationPoppedUp(true);
					}
					break;
				}
				case "MAIN": { //$NON-NLS-1$
					while (functionStack.size() > 0) {
						userFunction.doCancel();
						userFunction = (OnLinePresentationFunction) functionStack.removeLast();
					}
					if (userFunction.isSuspended()) { // Should always be true, but safety rules ...
						userFunction.unSuspend();
					}
					functionStack.addLast(userFunction);
					break;
				}
				case "LOGOFF": { //$NON-NLS-1$
					do {
						userFunction.doCancel();
						if (functionStack.size() > 0) {
							userFunction = (OnLinePresentationFunction) functionStack.removeLast();
						} else {
							userFunction = null;
						}
					} while (userFunction != null);
					break;
				}
				default: {
					userFunction.doActionLogic(action);
				}
			}
		} catch (ASPLException e1) {
			log(e1.getMessage(), e1);
			userFunction.setPresentationMessage(e1);
		}
		return userFunction;
	}

	/**
	 * A Request has been received that did not contain an Action= parameter. This means that either a new function is invoked, or else that we are returning from an invoked function to the function that was current when the invoke occurred.
	 * 
	 * @param req
	 *            The request
	 * @param applicationName
	 *            The application name
	 * @param functionName
	 *            The name of the user function in case it is a new invocation
	 * @param userSession
	 *            The user session in whose context
	 * @param userData
	 *            The userData object from the user session
	 * @return The presentation function that will write the output
	 * @throws ServletException
	 */
	private final OnLinePresentationFunction processInitOrResume (HttpServletRequest req, String applicationName, String functionName, UserSession userSession, UserData userData) throws ServletException {
		PresentationFunctionStack functionStack = userSession.getFunctionStack();

		// processing if there is no action : either function startup or function resume
		String returnPath = "", returnParms = ""; //$NON-NLS-1$ //$NON-NLS-2$
		if (functionStack.size() > 0) {
			OnLinePresentationFunction userFunction = (OnLinePresentationFunction) functionStack.getLast();
			if (userFunction.isSuspended()) {
				// This means that there has been a function atop of us, but since we are back again the top function now, we can conclude that that function
				// has done a doCancel(), thereby removing itself from the function stack, followed by a redirect to our URL. So we must now reactivate
				// (=unsuspend) ourself.  And we can immediately return, so the case of "return from invoked dialog" has been handled.
				userFunction.unSuspend();
				return userFunction;
			}

			// This means we have to suspend ourselves and invoke a new function atop of us. That function must be given our URL so that it knows where to return to
			// (using redirect) when it gets cancelled.
			returnPath = userFunction.getServletPath();
			returnParms = userFunction.getQueryString();

			// Suspend ourselves (iff the user function existed - which is not the case for the very first function invoked)
			userFunction.suspend();
		}

		// Create the presentationFunction object
		// String prefix = Architecture.getApplicationPackageName(applicationName);
		//		String userFunctionName = prefix + "." + functionName; //$NON-NLS-1$
		// Get the package name from the Servlet's Class. Each application should create its own Servlet than extends HTTPArchitectureServlet.
		String prefix = this.getClass().getPackage().getName();
		String userFunctionName = prefix + "." + functionName; //$NON-NLS-1$
		OnLinePresentationFunction newUserFunction;
		try {
			// newUserFunction = (OnLinePresentationFunction) Class.forName(userFunctionName).newInstance();
			newUserFunction = (OnLinePresentationFunction) getClass().getClassLoader().loadClass(userFunctionName).newInstance();
		} catch (InstantiationException e) {
			log(e.getMessage(), e);
			req.getSession().invalidate();
			Architecture.dropUserSession(userSession.getSessionID());
			throw new ServletException(e);
		} catch (IllegalAccessException e) {
			log(e.getMessage(), e);
			req.getSession().invalidate();
			Architecture.dropUserSession(userSession.getSessionID());
			throw new ServletException(e);
		} catch (ClassNotFoundException e) {
			log(e.getMessage(), e);
			req.getSession().invalidate();
			Architecture.dropUserSession(userSession.getSessionID());
			throw new ServletException(e);
		}

		// Update the stack - do this BEFORE initializing the function so that the stack already looks as if the function is the active one (and has already replaced its transferring callee).
		String functionStartupMode = req.getParameter("FUNCTIONSTARTUPMODE"); //$NON-NLS-1$
		if (functionStartupMode != null && functionStartupMode.equalsIgnoreCase("XFER") && functionStack.size() > 0) { //$NON-NLS-1$
			functionStack.removeLast();
		}
		functionStack.addLast(newUserFunction);

		// Initialize the presentationfunction
		newUserFunction.initialize(applicationName, req.getQueryString(), userData, returnPath, returnParms, functionName, getParameterValuesMap(req), new LinkedList<Part>(), this, req.getServletPath()); // Empty Parts list because we are only __initializing__ the function. Processing parts is after the user has pressed OK.

//		return newUserFunction;  the doInitialLogic that gets called inside the initialize(), can have decided to do a transfer to some other dialog.
		return (OnLinePresentationFunction) functionStack.getLast();
	}

	/**
	 * Does the returning to a previous function
	 * 
	 * @param req
	 *            The request that was handled and whose corresponding function is now returning
	 * @param resp
	 *            The response object through which the return must be handled
	 * @param userSession
	 *            The user session
	 * @param functionStack
	 *            The (user session's) function stack
	 * @return The user function that will receive control after the cancel, or null if there is no such function (meaning return is to the application main menu)
	 * @throws ServletException
	 */
	private final OnLinePresentationFunction processReturn (HttpServletRequest req, HttpServletResponse resp, UserSession userSession, PresentationFunctionStack functionStack) throws ServletException {
		OnLinePresentationFunction abandonedUserFunction = (OnLinePresentationFunction) functionStack.removeLast();
		OnLinePresentationFunction returnToUserFunction;
		if (functionStack.size() == 0) {
			returnToUserFunction = null;
		} else {
			returnToUserFunction = (OnLinePresentationFunction) functionStack.getLast();
			returnToUserFunction.saveInFunctionContext("__WRETURNFROM", abandonedUserFunction.getClass().getSimpleName()); //$NON-NLS-1$
			String backto = abandonedUserFunction.getReturnPath();
			String backparms = abandonedUserFunction.getReturnParms();
			if (backparms.indexOf("USERSESSIONID=") < 0) { //$NON-NLS-1$
				if (backparms.length() > 0) {
					backparms = backparms + "&USERSESSIONID=" + userSession.getSessionID() + "&USERSESSIONTOKEN=" + userSession.getSessionVerificationToken(); //$NON-NLS-1$ //$NON-NLS-2$
				} else {
					backparms = "USERSESSIONID=" + userSession.getSessionID() + "&USERSESSIONTOKEN=" + userSession.getSessionVerificationToken(); //$NON-NLS-1$ //$NON-NLS-2$
				}
			}
			String erdurl = resp.encodeRedirectURL("/" + backto + "?" + backparms); //$NON-NLS-1$ //$NON-NLS-2$
			try {
				resp.sendRedirect(erdurl);
			} catch (IOException e2) {
				log(e2.getMessage(), e2);
				req.getSession().invalidate();
				Architecture.dropUserSession(userSession.getSessionID());
				throw new ServletException(e2.getMessage());
			}
		}
		return returnToUserFunction;
	}

	/**
	 * @see javax.servlet.http.HttpServlet#doGet(javax.servlet.http.HttpServletRequest, javax.servlet.http.HttpServletResponse)
	 */
	public final void doGet (HttpServletRequest req, HttpServletResponse resp) throws ServletException {
		process(req, resp);
	}

	/**
	 * @see javax.servlet.http.HttpServlet#doPost(javax.servlet.http.HttpServletRequest, javax.servlet.http.HttpServletResponse)
	 */
	public final void doPost (HttpServletRequest req, HttpServletResponse resp) throws ServletException {
		process(req, resp);
	}

	/**
	 * @see javax.servlet.GenericServlet#getServletInfo()
	 */
	public final String getServletInfo ( ) {
		return this.getClass().getName() + " - " + this.toString(); //$NON-NLS-1$
	}

	/**
	 * Initializes the Servlet. The ASPL architecture is instantiated, and a reference to it is kept as long as the servlet object is kept by the Application server and the JVM. ServerInfo is obtained from the servlet context and copied into the architecture object for future reference by the
	 * 
	 * @see javax.servlet.GenericServlet#init()
	 */
	public void init ( ) throws ServletException {
		super.init();
		architecture = Architecture.getInstance();

		final ServletContext servletContext = getServletContext();
		// inject any properties configured in an 'application.properties' file in a subdirectory (named after the context) of the user.home directory, into the ServletContext object
		try (FileInputStream fis = new FileInputStream(System.getProperty("user.home") + File.separatorChar + servletContext.getContextPath().substring(1) + File.separatorChar + "application.properties");) {//$NON-NLS-1$//$NON-NLS-2$
			final Properties locallyConfiguredApplicationProperties = new Properties();
			locallyConfiguredApplicationProperties.load(fis);
			for (Entry<Object, Object> lce : locallyConfiguredApplicationProperties.entrySet()) {
				servletContext.setInitParameter((String) lce.getKey(), (String) lce.getValue());
			}
		} catch (FileNotFoundException e1) {

		} catch (IOException e1) {
			log(e1.getMessage(), e1);
		}
	}
}